@@ -1,11 +1,10 @@
 /* -*- c-file-style: "ruby"; indent-tabs-mode: nil -*- */
 /*
- *  Copyright (C) 2011-2015  Ruby-GNOME2 Project Team
- *  Copyright (C) 2002-2004  Ruby-GNOME2 Project Team
+ *  Copyright (C) 2002-2021  Ruby-GNOME Project Team
  *  Copyright (C) 2002-2003  Masahiro Sakai
- *  Copyright (C) 1998-2000 Yukihiro Matsumoto,
- *                          Daisuke Kanda,
- *                          Hiroshi Igarashi
+ *  Copyright (C) 1998-2000  Yukihiro Matsumoto,
+ *                           Daisuke Kanda,
+ *                           Hiroshi Igarashi
  *
  *  This library is free software; you can redistribute it and/or
  *  modify it under the terms of the GNU Lesser General Public
@@ -31,6 +30,12 @@ VALUE RG_TARGET_NAMESPACE;
 static VALUE eNoPropertyError;
 static GQuark RUBY_GOBJECT_OBJ_KEY;
 
+gboolean
+rbg_is_object(VALUE object)
+{
+    return RVAL2CBOOL(rb_obj_is_kind_of(object, RG_TARGET_NAMESPACE));
+}
+
 /* deperecated */
 void
 rbgobj_add_abstract_but_create_instance_class(G_GNUC_UNUSED GType gtype)
@@ -43,7 +48,8 @@ weak_notify(gpointer data, G_GNUC_UNUSED GObject *where_the_object_was)
     gobj_holder *holder = data;
 
     rbgobj_instance_call_cinfo_free(holder->gobj);
-    rbgobj_invalidate_relatives(holder->self);
+    g_hash_table_unref(holder->rb_relatives);
+    holder->rb_relatives = NULL;
     holder->destroyed = TRUE;
 
     g_object_unref(holder->gobj);
@@ -51,10 +57,26 @@ weak_notify(gpointer data, G_GNUC_UNUSED GObject *where_the_object_was)
 }
 
 static void
-holder_mark(gobj_holder *holder)
+holder_relatives_mark(G_GNUC_UNUSED gpointer key,
+                      gpointer value,
+                      G_GNUC_UNUSED gpointer user_data)
 {
-    if (holder->gobj && !holder->destroyed)
-        rbgobj_instance_call_cinfo_mark(holder->gobj);
+    VALUE rb_relative = (VALUE)value;
+    rb_gc_mark(rb_relative);
+}
+
+static void
+holder_mark(void *data)
+{
+    gobj_holder *holder = data;
+
+    if (!holder->gobj)
+        return;
+    if (holder->destroyed)
+        return;
+
+    rbgobj_instance_call_cinfo_mark(holder->gobj);
+    g_hash_table_foreach(holder->rb_relatives, holder_relatives_mark, NULL);
 }
 
 static void
@@ -71,23 +93,92 @@ holder_unref(gobj_holder *holder)
 }
 
 static void
-holder_free(gobj_holder *holder)
+holder_free(void *data)
 {
+    gobj_holder *holder = data;
     holder_unref(holder);
     xfree(holder);
 }
 
-static VALUE
-gobj_s_allocate(VALUE klass)
+static const rb_data_type_t rg_glib_object_type = {
+    "GLib::Object",
+    {
+        holder_mark,
+        holder_free,
+    },
+    NULL,
+    NULL,
+    RUBY_TYPED_FREE_IMMEDIATELY,
+};
+
+void
+rbgobj_object_add_relative(VALUE rb_gobject, VALUE rb_relative)
+{
+    gobj_holder *holder;
+    TypedData_Get_Struct(rb_gobject,
+                         gobj_holder,
+                         &rg_glib_object_type,
+                         holder);
+    if (holder->rb_relatives) {
+        g_hash_table_insert(holder->rb_relatives,
+                            (gpointer)(rb_relative),
+                            (gpointer)(rb_relative));
+    }
+}
+
+void
+rbgobj_object_remove_relative(VALUE rb_gobject, VALUE rb_relative)
+{
+    gobj_holder *holder;
+    TypedData_Get_Struct(rb_gobject,
+                         gobj_holder,
+                         &rg_glib_object_type,
+                         holder);
+    if (holder->rb_relatives) {
+        g_hash_table_remove(holder->rb_relatives,
+                            (gpointer)(rb_relative));
+    }
+}
+
+static gboolean
+rbgobj_object_remove_relatives_body(G_GNUC_UNUSED gpointer key,
+                                    gpointer value,
+                                    gpointer user_data)
+{
+    VALUE rb_relative = (VALUE)value;
+    VALUE rb_relative_class = (VALUE)user_data;
+
+    return RVAL2CBOOL(rb_obj_is_kind_of(rb_relative, rb_relative_class));
+}
+
+void
+rbgobj_object_remove_relatives(VALUE rb_gobject, VALUE rb_relative_class)
+{
+    gobj_holder *holder;
+    TypedData_Get_Struct(rb_gobject,
+                         gobj_holder,
+                         &rg_glib_object_type,
+                         holder);
+    g_hash_table_foreach_remove(holder->rb_relatives,
+                                rbgobj_object_remove_relatives_body,
+                                (gpointer)(rb_relative_class));
+}
+
+VALUE
+rbgobj_object_alloc_func(VALUE klass)
 {
     gobj_holder* holder;
     VALUE result;
 
-    result = Data_Make_Struct(klass, gobj_holder, holder_mark, holder_free, holder);
+    result = TypedData_Make_Struct(klass,
+                                   gobj_holder,
+                                   &rg_glib_object_type,
+                                   holder);
     holder->self  = result;
     holder->gobj  = NULL;
     holder->cinfo = NULL;
     holder->destroyed = FALSE;
+    holder->rb_relatives = g_hash_table_new(g_direct_hash, g_direct_equal);
 
     return result;
 }
@@ -96,7 +187,7 @@ gobj_s_allocate(VALUE klass)
 VALUE
 rbgobj_create_object(VALUE klass)
 {
-    return gobj_s_allocate(klass);
+    return rbgobj_object_alloc_func(klass);
 }
 
 void
@@ -105,7 +196,7 @@ rbgobj_gobject_initialize(VALUE obj, gpointer cobj)
     gobj_holder* holder = g_object_get_qdata((GObject*)cobj, RUBY_GOBJECT_OBJ_KEY);
     if (holder)
         rb_raise(rb_eRuntimeError, "ruby wrapper for this GObject* already exists.");
-    Data_Get_Struct(obj, gobj_holder, holder);
+    TypedData_Get_Struct(obj, gobj_holder, &rg_glib_object_type, holder);
     holder->cinfo = RVAL2CINFO(obj);
     holder->gobj  = (GObject*)cobj;
     holder->destroyed = FALSE;
@@ -135,7 +226,7 @@ rbgobj_get_ruby_object_from_gobject(GObject* gobj, gboolean alloc)
     } else if (alloc) {
         VALUE obj;
 
-        obj = gobj_s_allocate(GTYPE2CLASS(G_OBJECT_TYPE(gobj)));
+        obj = rbgobj_object_alloc_func(GTYPE2CLASS(G_OBJECT_TYPE(gobj)));
         gobj = g_object_ref(gobj);
         rbgobj_gobject_initialize(obj, (gpointer)gobj);
         return obj;
@@ -149,10 +240,7 @@ rbgobj_get_gobject(VALUE obj)
 {
     gobj_holder* holder;
 
-    if (!RVAL2CBOOL(rb_obj_is_kind_of(obj, GTYPE2CLASS(G_TYPE_OBJECT))))
-        rb_raise(rb_eTypeError, "not a GLib::Object");
-
-    Data_Get_Struct(obj, gobj_holder, holder);
+    TypedData_Get_Struct(obj, gobj_holder, &rg_glib_object_type, holder);
 
     if (holder->destroyed)
         rb_raise(rb_eTypeError, "destroyed GLib::Object");
@@ -232,8 +320,14 @@ struct param_setup_arg {
 };
 
 static VALUE
-_params_setup(VALUE arg, struct param_setup_arg *param_setup_arg)
+_params_setup(VALUE arg,
+              VALUE rb_param_setup_arg,
+              G_GNUC_UNUSED int argc,
+              G_GNUC_UNUSED const VALUE *argv,
+              G_GNUC_UNUSED VALUE block)
 {
+    struct param_setup_arg *param_setup_arg =
+        (struct param_setup_arg *)rb_param_setup_arg;
     guint index;
     VALUE name, val;
     GParamSpec* pspec;
@@ -267,16 +361,18 @@ _params_setup(VALUE arg, struct param_setup_arg *param_setup_arg)
 }
 
 static VALUE
-gobj_new_body(struct param_setup_arg* arg)
+gobj_new_body(VALUE rb_arg)
 {
+    struct param_setup_arg *arg = (struct param_setup_arg *)rb_arg;
     rb_iterate(rb_each, (VALUE)arg->params_hash, _params_setup, (VALUE)arg);
     return (VALUE)g_object_newv(G_TYPE_FROM_CLASS(arg->gclass),
                                 arg->param_size, arg->params);
 }
 
 static VALUE
-gobj_new_ensure(struct param_setup_arg* arg)
+gobj_new_ensure(VALUE rb_arg)
 {
+    struct param_setup_arg *arg = (struct param_setup_arg *)rb_arg;
     guint i;
     g_type_class_unref(arg->gclass);
     for (i = 0; i < arg->param_size; i++) {
@@ -299,10 +395,10 @@ rbgobj_gobject_new(GType gtype, VALUE params_hash)
     if (NIL_P(params_hash)) {
         result = g_object_newv(gtype, 0, NULL);
     } else {
-        size_t param_size;
+        guint param_size;
         struct param_setup_arg arg;
 
-        param_size = NUM2INT(rb_funcall(params_hash, rb_intern("length"), 0));
+        param_size = NUM2UINT(rb_funcall(params_hash, rb_intern("length"), 0));
 
         arg.param_size = param_size;
         arg.gclass = G_OBJECT_CLASS(g_type_class_ref(gtype));
@@ -399,140 +495,146 @@ gobj_s_properties(int argc, VALUE* argv, VALUE self)
     return ary;
 }
 
-static VALUE type_to_prop_setter_table;
-static VALUE type_to_prop_getter_table;
+static GHashTable *rbg_type_to_prop_setter_tables;
+static GMutex rbg_type_to_prop_setter_tables_mutex;
+static GHashTable *rbg_type_to_prop_getter_tables;
+static GMutex rbg_type_to_prop_getter_tables_mutex;
 
-void
-rbgobj_register_property_setter(GType gtype, const char *name, RValueToGValueFunc func)
+static void
+rbg_register_property_accessor(GHashTable *tables,
+                               GMutex *mutex,
+                               GType gtype,
+                               const char *name,
+                               gpointer accessor)
 {
-    GObjectClass* oclass;
-    GParamSpec* pspec;
+    g_mutex_lock(mutex);
 
-    VALUE table = rb_hash_aref(type_to_prop_setter_table, INT2FIX(gtype));
-    if (NIL_P(table)){
-        table = rb_hash_new();
-        rb_hash_aset(type_to_prop_setter_table, INT2FIX(gtype), table);
+    GHashTable *table = g_hash_table_lookup(tables, GUINT_TO_POINTER(gtype));
+    if (!table) {
+        table = g_hash_table_new(g_str_hash, g_str_equal);
+        g_hash_table_insert(tables, GUINT_TO_POINTER(gtype), table);
     }
 
-    oclass = g_type_class_ref(gtype);
-    pspec = g_object_class_find_property(oclass, name);
+    GObjectClass *gclass = g_type_class_ref(gtype);
+    GParamSpec *pspec = g_object_class_find_property(gclass, name);
+    g_hash_table_insert(table,
+                        (gchar *)g_param_spec_get_name(pspec),
+                        accessor);
+    g_type_class_unref(gclass);
 
-    rb_hash_aset(table, CSTR2RVAL(g_param_spec_get_name(pspec)),
-                 Data_Wrap_Struct(rb_cData, NULL, NULL, func));
+    g_mutex_unlock(mutex);
+}
 
-    g_type_class_unref(oclass);
+static gpointer
+rbg_get_property_accessor(GHashTable *tables,
+                          GMutex *mutex,
+                          GType gtype,
+                          const char *name)
+{
+    g_mutex_lock(mutex);
+
+    gpointer accessor = NULL;
+    GHashTable *table = g_hash_table_lookup(tables, GUINT_TO_POINTER(gtype));
+    if (table) {
+        GObjectClass *gclass = g_type_class_ref(gtype);
+        GParamSpec *pspec = g_object_class_find_property(gclass, name);
+        accessor = g_hash_table_lookup(table, g_param_spec_get_name(pspec));
+        g_type_class_unref(gclass);
+    }
+
+    g_mutex_unlock(mutex);
+
+    return accessor;
+}
+
+void
+rbgobj_register_property_setter(GType gtype,
+                                const char *name,
+                                RValueToGValueFunc func)
+{
+    rbg_register_property_accessor(rbg_type_to_prop_setter_tables,
+                                   &rbg_type_to_prop_setter_tables_mutex,
+                                   gtype,
+                                   name,
+                                   func);
 }
 
 void
 rbgobj_register_property_getter(GType gtype, const char *name, GValueToRValueFunc func)
 {
-    GObjectClass* oclass;
-    GParamSpec* pspec;
-
-    VALUE table = rb_hash_aref(type_to_prop_getter_table, INT2FIX(gtype));
-    if (NIL_P(table)){
-        table = rb_hash_new();
-        rb_hash_aset(type_to_prop_getter_table, INT2FIX(gtype), table);
-    }
-
-    oclass = g_type_class_ref(gtype);
-    pspec = g_object_class_find_property(oclass, name);
-
-    rb_hash_aset(table, CSTR2RVAL(g_param_spec_get_name(pspec)),
-                 Data_Wrap_Struct(rb_cData, NULL, NULL, func));
-
-    g_type_class_unref(oclass);
+    rbg_register_property_accessor(rbg_type_to_prop_getter_tables,
+                                   &rbg_type_to_prop_getter_tables_mutex,
+                                   gtype,
+                                   name,
+                                   func);
 }
 
 static VALUE
 rg_set_property(VALUE self, VALUE prop_name, VALUE val)
 {
-    GParamSpec* pspec;
     const char* name;
-
     if (SYMBOL_P(prop_name))
         name = rb_id2name(SYM2ID(prop_name));
     else
         name = StringValuePtr(prop_name);
 
-    pspec = g_object_class_find_property(G_OBJECT_GET_CLASS(RVAL2GOBJ(self)),
-                                         name);
-
+    GParamSpec* pspec =
+        g_object_class_find_property(G_OBJECT_GET_CLASS(RVAL2GOBJ(self)),
+                                     name);
     if (!pspec)
         rb_raise(eNoPropertyError, "No such property: %s", name);
-    else {
-        // FIXME: use rb_ensure to call g_value_unset()
-        RValueToGValueFunc setter = NULL;
-        GValue gval = G_VALUE_INIT;
 
-        g_value_init(&gval, G_PARAM_SPEC_VALUE_TYPE(pspec));
-
-        {
-            VALUE table = rb_hash_aref(type_to_prop_setter_table,
-                                       INT2FIX(pspec->owner_type));
-            if (!NIL_P(table)){
-                VALUE obj = rb_hash_aref(table, CSTR2RVAL(g_param_spec_get_name(pspec)));
-                if (!NIL_P(obj))
-                    Data_Get_Struct(obj, void, setter);
-            }
-        }
-        if (setter) {
-            setter(val, &gval);
-        } else {
-            rbgobj_rvalue_to_gvalue(val, &gval);
-        }
-
-        g_object_set_property(RVAL2GOBJ(self), name, &gval);
-        g_value_unset(&gval);
-
-        G_CHILD_SET(self, rb_intern(name), val);
-
-        return self;
+    RValueToGValueFunc setter =
+        rbg_get_property_accessor(rbg_type_to_prop_setter_tables,
+                                  &rbg_type_to_prop_setter_tables_mutex,
+                                  pspec->owner_type,
+                                  name);
+    // FIXME: use rb_ensure to call g_value_unset()
+    GValue gval = G_VALUE_INIT;
+    g_value_init(&gval, G_PARAM_SPEC_VALUE_TYPE(pspec));
+    if (setter) {
+        setter(val, &gval);
+    } else {
+        rbgobj_rvalue_to_gvalue(val, &gval);
     }
+    g_object_set_property(RVAL2GOBJ(self), name, &gval);
+    g_value_unset(&gval);
+
+    G_CHILD_SET(self, rb_intern(name), val);
+
+    return self;
 }
 
 static VALUE
 rg_get_property(VALUE self, VALUE prop_name)
 {
-    GParamSpec* pspec;
     const char* name;
-
     if (SYMBOL_P(prop_name))
         name = rb_id2name(SYM2ID(prop_name));
     else
         name = StringValuePtr(prop_name);
 
-    pspec = g_object_class_find_property(G_OBJECT_GET_CLASS(RVAL2GOBJ(self)),
-                                         name);
-
+    GParamSpec *pspec =
+        g_object_class_find_property(G_OBJECT_GET_CLASS(RVAL2GOBJ(self)),
+                                     name);
     if (!pspec)
         rb_raise(eNoPropertyError, "No such property: %s", name);
-    else {
-        // FIXME: use rb_ensure to call g_value_unset()
-        GValueToRValueFunc getter = NULL;
-        GValue gval = G_VALUE_INIT;
-        VALUE ret;
 
-        {
-            VALUE table = rb_hash_aref(type_to_prop_getter_table,
-                                       INT2FIX(pspec->owner_type));
-            if (!NIL_P(table)){
-                VALUE obj = rb_hash_aref(table, CSTR2RVAL(g_param_spec_get_name(pspec)));
-                if (!NIL_P(obj))
-                    Data_Get_Struct(obj, void, getter);
-            }
-        }
+    GValueToRValueFunc getter =
+        rbg_get_property_accessor(rbg_type_to_prop_getter_tables,
+                                  &rbg_type_to_prop_getter_tables_mutex,
+                                  pspec->owner_type,
+                                  name);
+    // FIXME: use rb_ensure to call g_value_unset()
+    GValue gval = G_VALUE_INIT;
+    g_value_init(&gval, G_PARAM_SPEC_VALUE_TYPE(pspec));
+    g_object_get_property(RVAL2GOBJ(self), name, &gval);
+    VALUE ret = getter ? getter(&gval) : GVAL2RVAL(&gval);
+    g_value_unset(&gval);
 
-        g_value_init(&gval, G_PARAM_SPEC_VALUE_TYPE(pspec));
-        g_object_get_property(RVAL2GOBJ(self), name, &gval);
+    G_CHILD_SET(self, rb_intern(name), ret);
 
-        ret = getter ? getter(&gval) : GVAL2RVAL(&gval);
-        g_value_unset(&gval);
-
-        G_CHILD_SET(self, rb_intern(name), ret);
-
-        return ret;
-    }
+    return ret;
 }
 
 static VALUE rg_thaw_notify(VALUE self);
@@ -566,10 +668,7 @@ rg_destroyed_p(VALUE self)
 {
     gobj_holder* holder;
 
-    if (!RVAL2CBOOL(rb_obj_is_kind_of(self, GTYPE2CLASS(G_TYPE_OBJECT))))
-        rb_raise(rb_eTypeError, "not a GLib::Object");
-
-    Data_Get_Struct(self, gobj_holder, holder);
+    TypedData_Get_Struct(self, gobj_holder, &rg_glib_object_type, holder);
 
     return CBOOL2RVAL(holder->destroyed);
 }
@@ -582,7 +681,7 @@ rg_inspect(VALUE self)
     char *s;
     VALUE result;
 
-    Data_Get_Struct(self, gobj_holder, holder);
+    TypedData_Get_Struct(self, gobj_holder, &rg_glib_object_type, holder);
 
     class_name = rb_class2name(CLASS_OF(self));
     if (!holder->destroyed)
@@ -602,7 +701,7 @@ rg_unref(VALUE self)
 {
     gobj_holder* holder;
 
-    Data_Get_Struct(self, gobj_holder, holder);
+    TypedData_Get_Struct(self, gobj_holder, &rg_glib_object_type, holder);
 
     if (holder->destroyed)
         rb_raise(rb_eTypeError, "destroyed GLib::Object");
@@ -620,7 +719,6 @@ rg_type_name(VALUE self)
     return CSTR2RVAL(G_OBJECT_TYPE_NAME(RVAL2GOBJ(self)));
 }
 
-#if GLIB_CHECK_VERSION(2, 26, 0)
 typedef struct {
     VALUE transform_from_callback;
     VALUE transform_to_callback;
@@ -675,13 +773,6 @@ static void
 rg_destroy_bind_property_full_data(gpointer user_data)
 {
     RGBindPropertyCallbackData *data = (RGBindPropertyCallbackData *)user_data;
-
-    if (!NIL_P(data->transform_to_callback))
-        G_CHILD_REMOVE(data->self, data->transform_to_callback);
-
-    if (!NIL_P(data->transform_from_callback))
-        G_CHILD_REMOVE(data->self, data->transform_from_callback);
-
     xfree(data);
 }
 
@@ -704,6 +795,7 @@ rg_bind_property(gint argc, VALUE *argv, VALUE self)
     GBinding *binding;
     GBindingTransformFunc transform_to = NULL;
     GBindingTransformFunc transform_from = NULL;
+    VALUE rb_binding;
 
     rb_scan_args(argc, argv, "41", &rb_source_property, &rb_target,
                  &rb_target_property, &rb_flags, &rb_options);
@@ -720,18 +812,16 @@ rg_bind_property(gint argc, VALUE *argv, VALUE self)
     flags = RVAL2GBINDINGFLAGS(rb_flags);
 
     if (!NIL_P(rb_transform_to)) {
-        G_CHILD_ADD(self, rb_transform_to);
         transform_to = rg_bind_property_transform_to_callback;
     }
 
     if (!NIL_P(rb_transform_from)) {
-        G_CHILD_ADD(self, rb_transform_from);
         transform_from = rg_bind_property_transform_from_callback;
     }
 
     if (transform_to || transform_from) {
         RGBindPropertyCallbackData *data;
-        data = (RGBindPropertyCallbackData *)xmalloc(sizeof(RGBindPropertyCallbackData));
+        data = RB_ALLOC(RGBindPropertyCallbackData);
         data->self = self;
         data->transform_to_callback = rb_transform_to;
         data->transform_from_callback = rb_transform_from;
@@ -741,15 +831,22 @@ rg_bind_property(gint argc, VALUE *argv, VALUE self)
                                               transform_from,
                                               (gpointer)data,
                                               rg_destroy_bind_property_full_data);
+        rb_binding = GOBJ2RVAL(binding);
+        if (!NIL_P(rb_transform_to)) {
+            rbgobj_object_add_relative(rb_binding, rb_transform_to);
+        }
+        if (!NIL_P(rb_transform_from)) {
+            rbgobj_object_add_relative(rb_binding, rb_transform_from);
+        }
     } else {
         binding = g_object_bind_property(source, source_property,
                                          target, target_property,
                                          flags);
+        rb_binding = GOBJ2RVAL(binding);
     }
 
-    return GOBJ2RVAL(binding);
+    return rb_binding;
 }
-#endif
 
 static VALUE
 rg_initialize(int argc, VALUE *argv, VALUE self)
@@ -780,10 +877,22 @@ static VALUE
 gobj_ref_count(VALUE self)
 {
     gobj_holder* holder;
-    Data_Get_Struct(self, gobj_holder, holder);
+    TypedData_Get_Struct(self, gobj_holder, &rg_glib_object_type, holder);
     return INT2NUM(holder->gobj ? holder->gobj->ref_count : 0);
 }
 
+static VALUE
+rg_floating_p(VALUE self)
+{
+    gobj_holder* holder;
+    TypedData_Get_Struct(self, gobj_holder, &rg_glib_object_type, holder);
+    if (holder->gobj) {
+        return CBOOL2RVAL(g_object_is_floating(holder->gobj));
+    } else {
+        return Qfalse;
+    }
+}
+
 /**********************************************************************/
 
 static GQuark q_ruby_setter;
@@ -960,7 +1069,7 @@ Init_gobject_gobject(void)
 
     RUBY_GOBJECT_OBJ_KEY = g_quark_from_static_string("__ruby_gobject_object__");
 
-    rb_define_alloc_func(RG_TARGET_NAMESPACE, (VALUE(*)_((VALUE)))gobj_s_allocate);
+    rb_define_alloc_func(RG_TARGET_NAMESPACE, rbgobj_object_alloc_func);
     RG_DEF_SMETHOD_BANG(new, -1);
 
     rbg_define_singleton_method(RG_TARGET_NAMESPACE, "property", &gobj_s_property, 1);
@@ -979,22 +1088,29 @@ Init_gobject_gobject(void)
 
     RG_DEF_METHOD(initialize, -1);
     rbg_define_method(RG_TARGET_NAMESPACE, "ref_count", gobj_ref_count, 0); /* for debugging */
+    RG_DEF_METHOD_P(floating, 0); /* for debugging */
     RG_DEF_METHOD(unref, 0);
     RG_DEF_METHOD(inspect, 0);
     RG_DEF_METHOD(type_name, 0);
 
-#if GLIB_CHECK_VERSION(2, 26, 0)
     RG_DEF_METHOD(bind_property, -1);
     G_DEF_CLASS(G_TYPE_BINDING_FLAGS, "BindingFlags", mGLib);
-#endif
 
     eNoPropertyError = rb_define_class_under(mGLib, "NoPropertyError",
                                              rb_eNameError);
 
-    rb_global_variable(&type_to_prop_setter_table);
-    rb_global_variable(&type_to_prop_getter_table);
-    type_to_prop_setter_table = rb_hash_new();
-    type_to_prop_getter_table = rb_hash_new();
+    rbg_type_to_prop_setter_tables =
+        g_hash_table_new_full(g_direct_hash,
+                              g_direct_equal,
+                              NULL,
+                              (GDestroyNotify)g_hash_table_unref);
+    g_mutex_init(&rbg_type_to_prop_setter_tables_mutex);
+    rbg_type_to_prop_getter_tables =
+        g_hash_table_new_full(g_direct_hash,
+                              g_direct_equal,
+                              NULL,
+                              (GDestroyNotify)g_hash_table_unref);
+    g_mutex_init(&rbg_type_to_prop_getter_tables_mutex);
 
     /* subclass */
     RG_DEF_SMETHOD(type_register, -1);
